@@ -147,12 +147,12 @@ static void handle_pinch_end(struct wl_listener *listener, void *data) {
 }
 
 static void handle_switch_toggle(struct wl_listener *listener, void *data) {
-	struct roots_switch *lid_switch =
-		wl_container_of(listener, lid_switch, toggle);
-	struct roots_desktop *desktop = lid_switch->seat->input->server->desktop;
-	wlr_idle_notify_activity(desktop->idle, lid_switch->seat->seat);
+	struct roots_switch *switch_device =
+		wl_container_of(listener, switch_device, toggle);
+	struct roots_desktop *desktop = switch_device->seat->input->server->desktop;
+	wlr_idle_notify_activity(desktop->idle, switch_device->seat->seat);
 	struct wlr_event_switch_toggle *event = data;
-	roots_switch_handle_toggle(lid_switch, event);
+	roots_switch_handle_toggle(switch_device, event);
 }
 
 static void handle_touch_down(struct wl_listener *listener, void *data) {
@@ -603,10 +603,38 @@ static void roots_drag_icon_handle_destroy(struct wl_listener *listener,
 	free(icon);
 }
 
-static void roots_seat_handle_new_drag_icon(struct wl_listener *listener,
+static void roots_seat_handle_request_start_drag(struct wl_listener *listener,
 		void *data) {
-	struct roots_seat *seat = wl_container_of(listener, seat, new_drag_icon);
-	struct wlr_drag_icon *wlr_drag_icon = data;
+	struct roots_seat *seat =
+		wl_container_of(listener, seat, request_start_drag);
+	struct wlr_seat_request_start_drag_event *event = data;
+
+	if (wlr_seat_validate_pointer_grab_serial(seat->seat,
+			event->origin, event->serial)) {
+		wlr_seat_start_pointer_drag(seat->seat, event->drag, event->serial);
+		return;
+	}
+
+	struct wlr_touch_point *point;
+	if (wlr_seat_validate_touch_grab_serial(seat->seat,
+			event->origin, event->serial, &point)) {
+		wlr_seat_start_touch_drag(seat->seat, event->drag, event->serial, point);
+		return;
+	}
+
+	wlr_log(WLR_DEBUG, "Ignoring start_drag request: "
+		"could not validate pointer or touch serial %" PRIu32, event->serial);
+	wlr_data_source_destroy(event->drag->source);
+}
+
+static void roots_seat_handle_start_drag(struct wl_listener *listener,
+		void *data) {
+	struct roots_seat *seat = wl_container_of(listener, seat, start_drag);
+	struct wlr_drag *wlr_drag = data;
+	struct wlr_drag_icon *wlr_drag_icon = wlr_drag->icon;
+	if (wlr_drag_icon == NULL) {
+		return;
+	}
 
 	struct roots_drag_icon *icon = calloc(1, sizeof(struct roots_drag_icon));
 	if (icon == NULL) {
@@ -649,20 +677,27 @@ static void roots_seat_handle_request_set_primary_selection(
 void roots_drag_icon_update_position(struct roots_drag_icon *icon) {
 	roots_drag_icon_damage_whole(icon);
 
-	struct wlr_drag_icon *wlr_icon = icon->wlr_drag_icon;
 	struct roots_seat *seat = icon->seat;
-	struct wlr_cursor *cursor = seat->cursor->cursor;
-	if (wlr_icon->is_pointer) {
+	struct wlr_drag *wlr_drag = icon->wlr_drag_icon->drag;
+	assert(wlr_drag != NULL);
+
+	switch (seat->seat->drag->grab_type) {
+	case WLR_DRAG_GRAB_KEYBOARD:
+		assert(false);
+	case WLR_DRAG_GRAB_KEYBOARD_POINTER:;
+		struct wlr_cursor *cursor = seat->cursor->cursor;
 		icon->x = cursor->x;
 		icon->y = cursor->y;
-	} else {
+		break;
+	case WLR_DRAG_GRAB_KEYBOARD_TOUCH:;
 		struct wlr_touch_point *point =
-			wlr_seat_touch_get_point(seat->seat, wlr_icon->touch_id);
+			wlr_seat_touch_get_point(seat->seat, wlr_drag->touch_id);
 		if (point == NULL) {
 			return;
 		}
 		icon->x = seat->touch_x;
 		icon->y = seat->touch_y;
+		break;
 	}
 
 	roots_drag_icon_damage_whole(icon);
@@ -738,8 +773,11 @@ struct roots_seat *roots_seat_create(struct roots_input *input, char *name) {
 		roots_seat_handle_request_set_primary_selection;
 	wl_signal_add(&seat->seat->events.request_set_primary_selection,
 		&seat->request_set_primary_selection);
-	seat->new_drag_icon.notify = roots_seat_handle_new_drag_icon;
-	wl_signal_add(&seat->seat->events.new_drag_icon, &seat->new_drag_icon);
+	seat->request_start_drag.notify = roots_seat_handle_request_start_drag;
+	wl_signal_add(&seat->seat->events.request_start_drag,
+		&seat->request_start_drag);
+	seat->start_drag.notify = roots_seat_handle_start_drag;
+	wl_signal_add(&seat->seat->events.start_drag, &seat->start_drag);
 	seat->destroy.notify = roots_seat_handle_destroy;
 	wl_signal_add(&seat->seat->events.destroy, &seat->destroy);
 
@@ -839,13 +877,13 @@ static void seat_add_pointer(struct roots_seat *seat,
 }
 
 static void handle_switch_destroy(struct wl_listener *listener, void *data) {
-	struct roots_switch *lid_switch =
-		wl_container_of(listener, lid_switch, device_destroy);
-	struct roots_seat *seat = lid_switch->seat;
+	struct roots_switch *switch_device =
+		wl_container_of(listener, switch_device, device_destroy);
+	struct roots_seat *seat = switch_device->seat;
 
-	wl_list_remove(&lid_switch->link);
-	wl_list_remove(&lid_switch->device_destroy.link);
-	free(lid_switch);
+	wl_list_remove(&switch_device->link);
+	wl_list_remove(&switch_device->device_destroy.link);
+	free(switch_device);
 
 	seat_update_capabilities(seat);
 }
@@ -853,19 +891,19 @@ static void handle_switch_destroy(struct wl_listener *listener, void *data) {
 static void seat_add_switch(struct roots_seat *seat,
 		struct wlr_input_device *device) {
 	assert(device->type == WLR_INPUT_DEVICE_SWITCH);
-		struct roots_switch *lid_switch = calloc(1, sizeof(struct roots_switch));
-	if (!lid_switch) {
+	struct roots_switch *switch_device = calloc(1, sizeof(struct roots_switch));
+	if (!switch_device) {
 		wlr_log(WLR_ERROR, "could not allocate switch for seat");
 		return;
 	}
-	device->data = lid_switch;
-	lid_switch->device = device;
-	lid_switch->seat = seat;
-	wl_list_insert(&seat->switches, &lid_switch->link);
-	lid_switch->device_destroy.notify = handle_switch_destroy;
+	device->data = switch_device;
+	switch_device->device = device;
+	switch_device->seat = seat;
+	wl_list_insert(&seat->switches, &switch_device->link);
+	switch_device->device_destroy.notify = handle_switch_destroy;
 
-	lid_switch->toggle.notify = handle_switch_toggle;
-	wl_signal_add(&lid_switch->device->lid_switch->events.toggle, &lid_switch->toggle);
+	switch_device->toggle.notify = handle_switch_toggle;
+	wl_signal_add(&switch_device->device->switch_device->events.toggle, &switch_device->toggle);
 }
 
 static void handle_touch_destroy(struct wl_listener *listener, void *data) {
@@ -1293,9 +1331,12 @@ void roots_seat_set_focus(struct roots_seat *seat, struct roots_view *view) {
 	bool unfullscreen = true;
 
 #if WLR_HAS_XWAYLAND
-	if (view && view->type == ROOTS_XWAYLAND_VIEW &&
-			view->xwayland_surface->override_redirect) {
-		unfullscreen = false;
+	if (view && view->type == ROOTS_XWAYLAND_VIEW) {
+		struct roots_xwayland_surface *xwayland_surface =
+			roots_xwayland_surface_from_view(view);
+		if (xwayland_surface->xwayland_surface->override_redirect) {
+			unfullscreen = false;
+		}
 	}
 #endif
 
@@ -1322,10 +1363,13 @@ void roots_seat_set_focus(struct roots_seat *seat, struct roots_view *view) {
 	}
 
 #if WLR_HAS_XWAYLAND
-	if (view && view->type == ROOTS_XWAYLAND_VIEW &&
-			!wlr_xwayland_or_surface_wants_focus(
-				view->xwayland_surface)) {
-		return;
+	if (view && view->type == ROOTS_XWAYLAND_VIEW) {
+		struct roots_xwayland_surface *xwayland_surface =
+			roots_xwayland_surface_from_view(view);
+		if (!wlr_xwayland_or_surface_wants_focus(
+				xwayland_surface->xwayland_surface)) {
+			return;
+		}
 	}
 #endif
 	struct roots_seat_view *seat_view = NULL;
