@@ -29,6 +29,11 @@ static void xdg_surface_configure_destroy(
 void unmap_xdg_surface(struct wlr_xdg_surface *surface) {
 	assert(surface->role != WLR_XDG_SURFACE_ROLE_NONE);
 
+	struct wlr_xdg_popup *popup, *popup_tmp;
+	wl_list_for_each_safe(popup, popup_tmp, &surface->popups, link) {
+		wlr_xdg_popup_destroy(popup->base);
+	}
+
 	// TODO: probably need to ungrab before this event
 	if (surface->mapped) {
 		wlr_signal_emit_safe(&surface->events.unmap, surface);
@@ -454,6 +459,11 @@ void reset_xdg_surface(struct wlr_xdg_surface *xdg_surface) {
 		wl_resource_set_user_data(xdg_surface->toplevel->resource, NULL);
 		xdg_surface->toplevel->resource = NULL;
 
+		if (xdg_surface->toplevel->client_pending.fullscreen_output) {
+			struct wlr_xdg_toplevel_state *client_pending =
+				&xdg_surface->toplevel->client_pending;
+			wl_list_remove(&client_pending->fullscreen_output_destroy.link);
+		}
 		free(xdg_surface->toplevel);
 		xdg_surface->toplevel = NULL;
 		break;
@@ -477,12 +487,6 @@ void reset_xdg_surface(struct wlr_xdg_surface *xdg_surface) {
 void destroy_xdg_surface(struct wlr_xdg_surface *surface) {
 	reset_xdg_surface(surface);
 
-	struct wlr_xdg_popup *popup_state, *next;
-	wl_list_for_each_safe(popup_state, next, &surface->popups, link) {
-		xdg_popup_send_popup_done(popup_state->resource);
-		destroy_xdg_popup(popup_state->base);
-	}
-
 	wl_resource_set_user_data(surface->resource, NULL);
 	surface->surface->role_data = NULL;
 
@@ -524,9 +528,20 @@ void wlr_xdg_toplevel_send_close(struct wlr_xdg_surface *surface) {
 }
 
 void wlr_xdg_popup_destroy(struct wlr_xdg_surface *surface) {
+	if (surface == NULL) {
+		return;
+	}
 	assert(surface->popup);
 	assert(surface->role == WLR_XDG_SURFACE_ROLE_POPUP);
+
+	struct wlr_xdg_popup *popup, *popup_tmp;
+	wl_list_for_each_safe(popup, popup_tmp, &surface->popups, link) {
+		wlr_xdg_popup_destroy(popup->base);
+	}
+
 	xdg_popup_send_popup_done(surface->popup->resource);
+	wl_resource_set_user_data(surface->popup->resource, NULL);
+	reset_xdg_surface(surface);
 }
 
 static void xdg_popup_get_position(struct wlr_xdg_popup *popup,
